from collections.abc import Generator
from operator import methodcaller

import pytest
from pytest import Config, FixtureRequest, MonkeyPatch

import urllib3
from pytest_urllib3._mock import Urllib3Mock
from pytest_urllib3._options import _Urllib3MockOptions
from pytest_urllib3._request import Request
from pytest_urllib3.version import __version__

__all__ = (
    "Urllib3Mock",
    "Request",
    "__version__",
)


def _build_full_url(pool: urllib3.HTTPConnectionPool, url: str) -> str:
    port_suffix = ""
    if pool.port is not None:
        is_default = (pool.scheme == "https" and pool.port == 443) or (
            pool.scheme == "http" and pool.port == 80
        )
        if not is_default:
            port_suffix = f":{pool.port}"
    return f"{pool.scheme}://{pool.host}{port_suffix}{url}"


@pytest.fixture
def urllib3_mock(
    monkeypatch: MonkeyPatch,
    request: FixtureRequest,
) -> Generator[Urllib3Mock, None, None]:
    options = {}
    for marker in request.node.iter_markers("urllib3_mock"):
        options = marker.kwargs | options
    __tracebackhide__ = methodcaller("errisinstance", TypeError)
    options = _Urllib3MockOptions(**options)

    mock = Urllib3Mock(options)

    real_urlopen = urllib3.HTTPConnectionPool.urlopen

    def mocked_urlopen(pool, method, url, body=None, headers=None, *args, **kwargs):
        full_url = _build_full_url(pool, url)
        if isinstance(body, str):
            body = body.encode("utf-8")
        req = Request(
            method=method.upper(),
            url=full_url,
            body=body,
            headers=dict(headers) if headers else {},
        )
        if options.should_mock(req):
            return mock._handle_request(req)
        return real_urlopen(pool, method, url, body, headers, *args, **kwargs)

    monkeypatch.setattr(urllib3.HTTPConnectionPool, "urlopen", mocked_urlopen)

    yield mock
    try:
        mock._assert_options()
    finally:
        mock.reset()


def pytest_configure(config: Config) -> None:
    config.addinivalue_line(
        "markers",
        "urllib3_mock(*, assert_all_responses_were_requested=True, assert_all_requests_were_expected=True, can_send_already_matched_responses=False, should_mock=lambda request: True): Configure urllib3_mock fixture.",
    )



import json
import re
from typing import Any, Optional, Union
from urllib.parse import parse_qs, quote, urlparse

from pytest_urllib3._options import _Urllib3MockOptions
from pytest_urllib3._request import Request


def _url_match(
    url_to_match: Union[re.Pattern[str], str],
    received_url: str,
    params: Optional[dict[str, Union[str, list[str]]]],
) -> bool:
    if isinstance(url_to_match, re.Pattern):
        return url_to_match.match(received_url) is not None

    parsed_expected = urlparse(url_to_match)
    parsed_received = urlparse(received_url)

    # Compare query parameters apart as order should not matter
    if params is None:
        expected_params = _to_params_dict(parse_qs(parsed_expected.query))
    else:
        expected_params = params
    received_params = _to_params_dict(parse_qs(parsed_received.query))

    # Normalize paths: percent-encode any non-ASCII characters so that
    # user-provided Unicode paths (e.g. "/données") match urllib3's
    # percent-encoded form (e.g. "/donn%C3%A9es"), while keeping
    # already-encoded sequences (e.g. "%2F") intact.
    expected_no_query = parsed_expected._replace(
        query="",
        path=quote(parsed_expected.path or "/", safe="/:@!$&'()*+,;=-._~%"),
    )
    received_no_query = parsed_received._replace(
        query="",
        path=quote(parsed_received.path or "/", safe="/:@!$&'()*+,;=-._~%"),
    )

    return (received_params == expected_params) and (
        expected_no_query == received_no_query
    )


def _to_params_dict(
    qs: dict[str, list[str]],
) -> dict[str, Union[str, list[str]]]:
    """Convert parse_qs output to a dict where single-value params are strings."""
    return {
        key: values[0] if len(values) == 1 else values for key, values in qs.items()
    }


class _RequestMatcher:
    def __init__(
        self,
        options: _Urllib3MockOptions,
        url: Optional[Union[str, re.Pattern[str]]] = None,
        method: Optional[str] = None,
        match_headers: Optional[dict[str, Any]] = None,
        match_content: Optional[bytes] = None,
        match_json: Optional[Any] = None,
        match_params: Optional[dict[str, Union[str, list[str]]]] = None,
        is_optional: Optional[bool] = None,
        is_reusable: Optional[bool] = None,
    ):
        self._options = options
        self.nb_calls = 0
        self.url = url
        self.method = method.upper() if method else method
        self.headers = match_headers
        self.content = match_content
        self.json = match_json
        self.params = match_params
        self.is_optional = (
            not options.assert_all_responses_were_requested
            if is_optional is None
            else is_optional
        )
        self.is_reusable = (
            options.can_send_already_matched_responses
            if is_reusable is None
            else is_reusable
        )
        if self._is_matching_body_more_than_one_way():
            raise ValueError(
                "Only one way of matching against the body can be provided. Use match_json for JSON or match_content for raw bytes."
            )
        if self.params and not self.url:
            raise ValueError("URL must be provided when match_params is used.")
        if self.params is not None and isinstance(self.url, re.Pattern):
            raise ValueError("match_params cannot be used in addition to regex URL.")
        if self._is_matching_params_more_than_one_way():
            raise ValueError(
                "Provided URL must not contain any query parameter when match_params is used."
            )

    def _is_matching_body_more_than_one_way(self) -> bool:
        return self.content is not None and self.json is not None

    def _is_matching_params_more_than_one_way(self) -> bool:
        url_has_params = isinstance(self.url, str) and bool(urlparse(self.url).query)
        return self.params is not None and url_has_params

    def match(self, request: Request) -> bool:
        return (
            self._url_match(request)
            and self._method_match(request)
            and self._headers_match(request)
            and self._content_match(request)
        )

    def _url_match(self, request: Request) -> bool:
        if not self.url:
            return True
        return _url_match(self.url, request.url, self.params)

    def _method_match(self, request: Request) -> bool:
        if not self.method:
            return True
        return request.method == self.method

    def _headers_match(self, request: Request) -> bool:
        if not self.headers:
            return True
        lower_headers = {k.lower(): v for k, v in request.headers.items()}
        return all(
            lower_headers.get(name.lower()) == value
            for name, value in self.headers.items()
        )

    def _content_match(self, request: Request) -> bool:
        if self.content is not None:
            return request.content == self.content

        if self.json is not None:
            try:
                return json.loads(request.content.decode("utf-8")) == self.json
            except (json.JSONDecodeError, UnicodeDecodeError):
                return False

        return True

    def should_have_matched(self) -> bool:
        return not self.is_optional and not self.nb_calls

    def __str__(self) -> str:
        if self.is_reusable:
            matcher_description = f"Match {self.method or 'every'} request"
        else:
            matcher_description = "Already matched" if self.nb_calls else "Match"
            matcher_description += f" {self.method or 'any'} request"
        if self.url:
            matcher_description += f" on {self.url}"
        if extra_description := self._extra_description():
            matcher_description += f" with {extra_description}"
        return matcher_description

    def _extra_description(self) -> str:
        extra_description = []
        if self.params:
            extra_description.append(f"{self.params} query parameters")
        if self.headers:
            extra_description.append(f"{self.headers} headers")
        if self.content is not None:
            extra_description.append(f"{self.content} body")
        if self.json is not None:
            extra_description.append(f"{self.json} json body")
        return " and ".join(extra_description)



from pytest_urllib3._request import Request
from pytest_urllib3._request_matcher import _RequestMatcher


def describe_request(
    request: Request,
    matchers: list[_RequestMatcher],
) -> str:
    expected_headers = {
        header for matcher in matchers if matcher.headers for header in matcher.headers
    }
    expect_body = any(
        matcher.content is not None or matcher.json is not None for matcher in matchers
    )

    description = f"{request.method} request on {request.url}"

    extra = []
    if expected_headers:
        lower_expected = {h.lower() for h in expected_headers}
        present_headers = {
            name: value
            for name, value in request.headers.items()
            if name.lower() in lower_expected
        }
        extra.append(f"{present_headers} headers")
    if expect_body:
        extra.append(f"{request.content} body")
    if extra:
        description += f" with {' and '.join(extra)}"

    return description


def explain_no_response_found(
    request: Request,
    matchers: list[_RequestMatcher],
) -> str:
    message = f"No response can be found for {describe_request(request, matchers)}"

    already_matched = []
    unmatched = []
    for matcher in matchers:
        if matcher.nb_calls:
            already_matched.append(matcher)
        else:
            unmatched.append(matcher)

    matchers_description = "\n".join(
        [f"- {matcher}" for matcher in unmatched + already_matched]
    )
    if matchers_description:
        message += f" amongst:\n{matchers_description}"
        if any(not matcher.is_reusable for matcher in already_matched):
            message += "\n\nIf you wanted to reuse an already matched response instead of registering it again, refer to is_reusable=True or can_send_already_matched_responses option."

    return message



import copy
import http.client
import io
import json as json_module
from collections.abc import Callable
from typing import Any, NoReturn, Optional

import urllib3.exceptions

import urllib3
from pytest_urllib3._options import _Urllib3MockOptions
from pytest_urllib3._pretty_print import explain_no_response_found
from pytest_urllib3._request import Request
from pytest_urllib3._request_matcher import _RequestMatcher


def _build_response(
    status_code: int = 200,
    headers: Optional[dict[str, str]] = None,
    content: Optional[bytes] = None,
    text: Optional[str] = None,
    json: Any = None,
) -> urllib3.HTTPResponse:
    final_headers = dict(headers or {})

    if json is not None:
        body = json_module.dumps(json).encode("utf-8")
        final_headers.setdefault("Content-Type", "application/json")
    elif text is not None:
        body = text.encode("utf-8")
        final_headers.setdefault("Content-Type", "text/plain; charset=utf-8")
    elif content is not None:
        body = content
    else:
        body = b""

    final_headers.setdefault("Content-Length", str(len(body)))

    return urllib3.HTTPResponse(
        body=io.BytesIO(body),
        status=status_code,
        headers=final_headers,
        preload_content=False,
        decode_content=False,
        version=11,
        version_string="HTTP/1.1",
        reason=http.client.responses.get(status_code, "Unknown"),
    )


class Urllib3Mock:
    def __init__(self, options: _Urllib3MockOptions) -> None:
        self._options = options
        self._requests: list[Request] = []
        self._callbacks: list[
            tuple[_RequestMatcher, Callable[[Request], Optional[urllib3.HTTPResponse]]]
        ] = []
        self._requests_not_matched: list[Request] = []

    def add_response(
        self,
        status_code: int = 200,
        headers: Optional[dict[str, str]] = None,
        content: Optional[bytes] = None,
        text: Optional[str] = None,
        json: Any = None,
        **matchers: Any,
    ) -> None:
        """Queue a response to return when a matching request is made.

        :param status_code: HTTP status code (default 200).
        :param headers: Response headers dict.
        :param content: Response body as bytes.
        :param text: Response body as string (sets Content-Type: text/plain).
        :param json: Response body as JSON-serializable object (sets Content-Type: application/json).
        :param **matchers: Request matching criteria (url, method, match_headers, match_content, match_json, match_params, is_optional, is_reusable).
        """
        json = copy.deepcopy(json) if json is not None else None

        def response_callback(request: Request) -> urllib3.HTTPResponse:
            return _build_response(
                status_code=status_code,
                headers=headers,
                content=content,
                text=text,
                json=json,
            )

        self.add_callback(response_callback, **matchers)

    def add_callback(
        self,
        callback: Callable[[Request], Optional[urllib3.HTTPResponse]],
        **matchers: Any,
    ) -> None:
        """Register a callback to handle matching requests.

        :param callback: Callable receiving a Request, must return a urllib3.HTTPResponse (or None).
        :param **matchers: Request matching criteria.
        """
        self._callbacks.append((_RequestMatcher(self._options, **matchers), callback))

    def add_exception(self, exception: BaseException, **matchers: Any) -> None:
        """Raise an exception when a matching request is made.

        :param exception: The exception to raise.
        :param **matchers: Request matching criteria.
        """

        def exception_callback(request: Request) -> None:
            raise exception

        self.add_callback(exception_callback, **matchers)

    def _handle_request(self, request: Request) -> urllib3.HTTPResponse:
        self._requests.append(request)

        callback = self._get_callback(request)
        if callback:
            response = callback(request)
            if response:
                return response

        self._request_not_matched(request)

    def _request_not_matched(self, request: Request) -> NoReturn:
        self._requests_not_matched.append(request)
        matchers = [matcher for matcher, _ in self._callbacks]
        raise urllib3.exceptions.TimeoutError(
            explain_no_response_found(request, matchers)
        )

    def _get_callback(
        self, request: Request
    ) -> Optional[Callable[[Request], Optional[urllib3.HTTPResponse]]]:
        callbacks = [
            (matcher, callback)
            for matcher, callback in self._callbacks
            if matcher.match(request)
        ]

        if not callbacks:
            return None

        for matcher, callback in callbacks:
            if not matcher.nb_calls:
                matcher.nb_calls += 1
                return callback

        # All matching callbacks have been used; reuse the last one if allowed
        last_matcher, last_callback = callbacks[-1]
        if last_matcher.is_reusable:
            last_matcher.nb_calls += 1
            return last_callback

        return None

    def get_requests(self, **matchers: Any) -> list[Request]:
        """Return all captured requests matching the given criteria.

        :param **matchers: Request matching criteria (url, method, match_headers, match_content, match_json, match_params).
        """
        matcher = _RequestMatcher(self._options, **matchers)
        return [request for request in self._requests if matcher.match(request)]

    def get_request(self, **matchers: Any) -> Optional[Request]:
        """Return the single captured request matching the criteria, or None.

        :param **matchers: Request matching criteria.
        :raises AssertionError: If more than one request matches.
        """
        requests = self.get_requests(**matchers)
        assert len(requests) <= 1, (
            f"More than one request ({len(requests)}) matched, use get_requests instead or refine your filters."
        )
        return requests[0] if requests else None

    def reset(self) -> None:
        """Clear all registered responses, callbacks, and captured requests."""
        self._requests.clear()
        self._callbacks.clear()
        self._requests_not_matched.clear()

    def _assert_options(self) -> None:
        callbacks_not_executed = [
            matcher for matcher, _ in self._callbacks if matcher.should_have_matched()
        ]
        matchers_description = "\n".join(
            [f"- {matcher}" for matcher in callbacks_not_executed]
        )

        assert not callbacks_not_executed, (
            f"The following responses are mocked but not requested:\n{matchers_description}\n\nIf this is on purpose, use is_optional=True or assert_all_responses_were_requested=False option."
        )

        if self._options.assert_all_requests_were_expected:
            requests_description = "\n".join(
                [
                    f"- {request.method} request on {request.url}"
                    for request in self._requests_not_matched
                ]
            )
            assert not self._requests_not_matched, (
                f"The following requests were not expected:\n{requests_description}\n\nIf this is on purpose, use assert_all_requests_were_expected=False option."
            )



from dataclasses import dataclass, field
from typing import Optional


@dataclass
class Request:
    method: str
    url: str
    body: Optional[bytes] = None
    headers: dict = field(default_factory=dict)

    @property
    def content(self) -> bytes:
        if self.body is None:
            return b""
        return self.body
